# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""SQLAlchemy query builder."""

from pygasus.storage.query_builder import AbstractQueryBuilder


class SQLQueryBuilder(AbstractQueryBuilder):

    """Query builder for SQLAlchemy."""

    def _get_table(self, field):
        """Return the table for this model."""
        model = field.__model__
        model_name = getattr(
            model.__config__, "model_name", model.__name__.lower()
        )
        return self.storage_engine.tables[model_name]

    def eq(self, field, other):
        """Compare field to other."""
        table = self._get_table(field)
        return getattr(table.c, field.name) == other

    def ne(self, field, other):
        """Compare field to other."""
        table = self._get_table(field)
        return getattr(table.c, field.name) != other

    def lt(self, field, other):
        """Compare field to other."""
        table = self._get_table(field)
        return getattr(table.c, field.name) < other

    def le(self, field, other):
        """Compare field to other."""
        table = self._get_table(field)
        return getattr(table.c, field.name) <= other

    def gt(self, field, other):
        """Compare field to other."""
        table = self._get_table(field)
        return getattr(table.c, field.name) > other

    def ge(self, field, other):
        """Compare field to other."""
        table = self._get_table(field)
        return getattr(table.c, field.name) >= other

    def is_in(self, field, collection):
        """Filter fields with a value in a collection."""
        table = self._get_table(field)
        return getattr(table.c, field.name).in_(collection)

    def is_not_in(self, field, collection):
        """Filter fields with a value not in a collection."""
        table = self._get_table(field)
        return getattr(table.c, field.name).not_in(collection)

    def has(self, field, value):
        """Return models with the field having this value (flag)."""
        table = self._get_table(field)
        return getattr(table.c, field.name).op("&")(value.value) == value.value

    def has_not(self, field, value):
        """Return models without the field having this value (flag)."""
        table = self._get_table(field)
        return getattr(table.c, field.name).op("&")(value.value) != value.value



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

"""Helper functions for models."""

from typing import Any, Dict, Type, TYPE_CHECKING

if TYPE_CHECKING:
    from pygasus.model.base import Model  # pragma: no cover


def get_model_qualname(model: Type["Model"]) -> str:
    """Return the model's qualname with its module.

    Args:
        model (subclass of Model): the class.

    Returns:
        qualname (str): the qualname with its module.

    """
    return f"{model.__module__}.{model.__qualname__}"


def get_primary_keys(
    model: "Model",
    include_model_name: bool = False,
) -> Dict[str, Any]:
    """Get the primary keys of this object and return a dictionary.

    Args:
        model (Model): the model from which to retrieve primary key fields.
        include_model_name (bool): include the model name in the dictionary.

    Returns:
        A dictionary.  The first entry, `__name__`, is the model
        name, if `include_model_name` has been set to `True`.

    """
    keys = {}
    model_cls = type(model)
    if include_model_name:
        keys["__name__"] = get_model_qualname(model_cls)

    # Add the primary key fields and their values.
    for name, field in model.__fields__.items():
        if field.field_info.extra.get("primary_key", False):
            value = getattr(model, name, ...)
            if value is ...:
                raise ValueError(
                    f"the model of class {model_cls} should have "
                    f"a field named {name!r}, this is not the case"
                )
            keys[name] = value

    return keys



# Copyright (c) 2022, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

"""Field dynamic decorator to wrap pydantic.ModelField."""

from typing import Any, TYPE_CHECKING

from pydantic.main import ModelField

if TYPE_CHECKING:
    from pygasus.model.base import Model  # pragma: no cover
    from pygasus.storage.abc import AbstractStorageEngine  # pragma: no cover


class PygasusField:

    """Pygasus field, wrapping around a Pydantic ModelField."""

    def __init__(self, field: ModelField):
        self.__field__ = field
        self.__model__ = None
        self.__storage__ = None

    def __getattr__(self, key):
        if key.startswith("__") and key.endswith("__"):
            raise AttributeError(
                f"class {type(self).__name__} has no attribute {key!r}"
            )
        return getattr(self.__field__, key)

    def __eq__(self, other):
        return self.__storage__.query_builder.eq(self, other)

    def __ne__(self, other):
        return self.__storage__.query_builder.ne(self, other)

    def __lt__(self, other):
        return self.__storage__.query_builder.lt(self, other)

    def __le__(self, other):
        return self.__storage__.query_builder.le(self, other)

    def __gt__(self, other):
        return self.__storage__.query_builder.gt(self, other)

    def __ge__(self, other):
        return self.__storage__.query_builder.ge(self, other)

    def is_in(self, collection):
        """Return models with the field in this collection."""
        return self.__storage__.query_builder.is_in(self, collection)

    def is_not_in(self, collection):
        """Return models with the field not in this collection."""
        return self.__storage__.query_builder.is_not_in(self, collection)

    def has(self, value):
        """Return models with the field having this value (flag)."""
        return self.__storage__.query_builder.has(self, value)

    def has_not(self, value):
        """Return models without the field having this value (flag)."""
        return self.__storage__.query_builder.has_not(self, value)

    # Methods that can be overridden by subclasses.
    def bind(self, model: "Model", storage: "AbstractStorageEngine"):
        """Bind this field to the model and storage engine."""
        self.__model__ = model
        self.__storage__ = storage

    def validate_update(self, model: "Model", old_value: Any, new_value: Any):
        """Validate whether updating this field is possible.

        This method should raise an appropriate exception
        if the update isn't allowed for any reason.

        Args:
            model (Model): the model object to be updated.
            old_value (Any): the old value.
            new_value (Any): the new value which will be applied
                    if this method doesn't raise any exception.

        """

    def perform_update(self, model: "Model", old_value: Any, new_value: Any):
        """Update and propagate the update of this field.

        This method is used to propagate updates to linked fields,
        for instance, the value of updated linked fields through relations.

        Args:
            model (Model): the model object to be updated.
            old_value (Any): the old value.
            new_value (Any): the new value already set in the model.

        """

    def validate_delete(self, model: "Model"):
        """Validate whether deleting this field is possible.

        This method should raise an appropriate exception
        if the deletion isn't allowed for any reason.

        Args:
            model (Model): the model object to be updated.

        """

    def perform_delete(self, model: "Model"):
        """Delete and propagate if necessary.

        This method is used to propagate deletions to linked fields,
        for instance, the value of deleted linked fields through relations.

        Args:
            model (Model): the model object to be updated.

        """



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

"""Base Model for all Pygasus stored objects.

All data models destined to be stored in Pygasus should inherit,
directly or indirectly, from the `Model` class defined here.
Models are described using simple type annotation
and inherit from Pydantic `BaseModel`, so they share the features
of models defined in Pydantic.

```python
from typing import Optional

grom pygasus import Model

class User(Model):

    '''A User model.'''

    id: int
    name: str
    age: int
    gender: Optional[str]
```

Behind the scenes, a Pygasus model is meant to be light and mostly
independent from the storage utility.  It is linked to a repository,
which is defined bvy default (but you can easily override it).
This repository has access to the storage utility and can perform
more advanced queries.  For convenience's sake, the `__setattr__` method
of models is linked to the UPDATE operation on the repository, so that
modifying a model will update it in storage as well.  This is the only
convenience (creating a model as usual with `__init__` won't do
anything in regard to the storage).  Of course, you can still access
the repository and perform these operations:

```python
# This will perform an INSERT operation on the user repository.
new_user = User.repository.create(name="Vincent", age=33)
# At this point, the new user has an ID which is set by the storage utility.
print(new_user.id)
# And this will perform an UPDATE operation.
new_user.age = 21
# This will fetch the user with id of 5.
user_5 = User.repository.get(id=5)
```

"""

from importlib import import_module
from typing import Any, Optional, Type

from pydantic import BaseModel, Field, PrivateAttr  # noqa: F401
from pydantic.main import ModelMetaclass

from pygasus.field.base import PygasusField
from pygasus.model.collections import Sequence
from pygasus.model.decorators import LazyPropertyDescriptor, lazy_property
from pygasus.model.helpers import get_primary_keys
from pygasus.model.repository import Repository

MODELS = set()


class MetaModel(ModelMetaclass):

    """Metaclass for all models."""

    def __new__(cls, name, bases, attrs):
        cls = super().__new__(cls, name, bases, attrs)
        cls.__pygasus__ = {}
        if cls.__name__ != "Model":
            MODELS.add(cls)

            # Wrap fields.
            for name, field in cls.__fields__.items():
                pygasus_field = PygasusField(field)
                cls.__pygasus__[name] = pygasus_field
                setattr(cls, name, pygasus_field)
        return cls

    @property
    def repository_path(cls):
        """Return the path leading to the repository, if any."""
        path = cls.__config__.repository_path
        if path is not None:
            return path

        return path

    @lazy_property
    def repository(cls) -> Optional[Repository]:
        """Return the repository object."""
        return load_repository(cls.repository_path, cls)


class Model(BaseModel, metaclass=MetaModel):

    """Abstract class for all models."""

    _exists = PrivateAttr()

    class Config:
        extra = "forbid"
        keep_untouched = (LazyPropertyDescriptor,)
        repository_path = None
        unique = ()
        copy_on_model_validation = False
        arbitrary_types_allowed = True

    def __init__(self, **kwargs):
        self._exists = False
        super().__init__(**kwargs)
        self._exists = True

    def __getattribute__(self, attr: str) -> Any:
        """Get the attribute doing a storage query if necessary."""
        value = object.__getattribute__(self, attr)
        if attr.startswith("_") or not self._exists:
            return value

        field = self.__fields__.get(attr)
        if field is not None:
            o_type = field.outer_type_
            f_type = field.type_
            if (
                isinstance(f_type, type)
                and issubclass(f_type, Model)
                and o_type is Sequence[f_type]
            ):
                # This object has several related, query the storage if needed.
                value.load_from_storage()
        return value

    def __repr_args__(self):
        args = dict(self.__dict__)
        for key, value in args.items():
            if isinstance(value, Model):
                pkeys = get_primary_keys(value)
                args[key] = str(pkeys)

        return args.items()

    def __setattr__(self, key: str, value: Any):
        """Force an UPDATE operation on the storage."""
        old_value = getattr(self, key, ...)
        exists = getattr(self, "_exists", False)
        cls_attr = getattr(type(self), key, None)
        if isinstance(cls_attr, (property, LazyPropertyDescriptor)):
            object.__setattr__(self, key, value)
            return

        if exists and not key.startswith("_"):
            # Inform the repositories of a change.
            repository = type(self).repository
            if repository:
                repository.update(self, key, old_value, value)
        super().__setattr__(key, value)

    def __eq__(self, other):
        """Don't compare dict like Pydantic does, just compare field values."""
        if not isinstance(other, type(self)):
            return False

        common = True
        for name, field in self.__fields__.items():
            value1 = getattr(self, name, ...)
            value2 = getattr(other, name, ...)
            if isinstance(value1, Model) or isinstance(value2, Model):
                common = get_primary_keys(
                    value1, include_model_name=True
                ) == get_primary_keys(value2, include_model_name=True)
            elif value1 != value2:
                common = False

            if not common:
                break

        return common


def load_repository(path: str, model: Type[Model]) -> Repository:
    """Try to load and return the repository."""
    if path is None:
        return Repository(model)

    module = import_module(path)

    # Look into the path, there may not be more than one class,
    # this is the repository.
    repository_cls = None
    for name, value in module.__dict__.items():
        if name.startswith("_"):
            continue

        if isinstance(value, type) and issubclass(value, Repository):
            if value.__module__ != path:
                continue

            if repository_cls is not None:
                repository_cls = ...
                break
            else:
                repository_cls = value

    if repository_cls is None:
        raise ValueError(f"No repository could be found in {path}.")
    elif repository_cls is ...:
        raise ValueError(
            "More than one repository are present "
            f"in module {path}, not loading any."
        )
    else:
        return repository_cls(model)



# Copyright (c) 20201, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

"""Lazy property, to optimize getting and setting data.

The lazy property descriptor is used similarly to a property,
but it caches the data it retrieves the first time it's called
and then will only return this cached data, unless a setter
is called in the meantime.

"""

_MISSING = object()


class LazyPropertyDescriptor:

    """
    Delays loading of property until first access.

    Although extended, this was inspired by Evennia's utility
    (wwww.evennia.com), itself based on the iplementation in the
    werkzeug suite:
        http://werkzeug.pocoo.org/docs/utils/#werkzeug.utils.cached_property

    A lazy property should be used as a decorator over the getter method,
    just like a property.  The difference is that a lazy property will
    call the getter method only once, the first time for this object, and
    then cache the result for following queries.  This allows for fast-access
    to handlers that are not re-created each time the property is called:

        ```python
        class SomeTest(Model):
            @lazy_property
            def db(self):
                return AttributeHandler(self)

            @db.setter
            def db(self, handler):
                raise ValueError("you can't change that")
        ```

    Once initialized, the `AttributeHandler` will be available as a
    property "db" on the object.

    """

    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset
        self.memory = {}

    def __get__(self, instance, owner=None):
        if instance is None:
            return self

        # The value might be cached in `memory`
        try:
            identifier = hash(instance)
        except TypeError:
            identifier = None
            attr = self.fget.__name__
            cached_attr = f"_cached_{attr}"

        if identifier:
            value = self.memory.get(identifier, _MISSING)
        else:
            value = getattr(instance, cached_attr, _MISSING)

        if value is _MISSING:
            value = self.fget(instance)
            if identifier:
                self.memory[identifier] = value
            else:
                setattr(instance, cached_attr, value)

        return value

    def __set__(self, instance, value):
        if not self.fset:
            raise AttributeError("can't set attribute")

        try:
            identifier = hash(instance)
        except TypeError:
            identifier = None
            attr = self.fget.__name__
            cached_attr = f"_cached_{attr}"

        self.fset(instance, value)

        if identifier:
            self.memory[identifier] = value
        else:
            setattr(instance, cached_attr, value)

    def setter(self, func):
        self.fset = func
        return self


def lazy_property(func):
    return LazyPropertyDescriptor(func)



# Copyright (c) 2022, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

"""A one-to-one relationship as a Pygasus field."""

from typing import Any, TYPE_CHECKING

from pydantic.main import ModelField

from pygasus.field.base import PygasusField

if TYPE_CHECKING:
    from pygasus.model.base import Model  # pragma: no cover
    from pygasus.storage.abc import AbstractStorageEngine  # pragma: no cover


class One2OneField(PygasusField):

    """One-to-one field, wrapping around a Pydantic ModelField."""

    def __init__(self, field: ModelField):
        self.__field__ = field
        self.__model__ = None
        self.__storage__ = None
        self.__required__ = ...
        self.__back__ = ...

    def bind(self, model: "Model", storage: "AbstractStorageEngine"):
        """Bind this field to the model and storage engine."""
        field = self.__field__
        self.__model__ = model
        self.__storage__ = storage
        self.__required__ = field.required
        back = storage.get_back_field(model, field, field.type_)
        self.__back__ = getattr(field.type_, back.name)

    def validate_update(self, model: "Model", old_value: Any, new_value: Any):
        """Validate whether updating this field is possible.

        This method should raise an appropriate exception
        if the update isn't allowed for any reason.

        Args:
            model (Model): the model object to be updated.
            old_value (Any): the old value.
            new_value (Any): the new value which will be applied
                    if this method doesn't raise any exception.

        """
        if new_value is None and self.required:
            raise ValueError(
                f"this operation would set the field {self.name} to "
                f"None on {model}, which would break model consistency."
            )

        back = self.__back__
        old = getattr(model, self.name, None)
        if old and back.required:
            raise ValueError(
                f"this operation would set the field {back.name} to "
                f"None on {old}, which would break model consistency."
            )

        # If the new value had a set field before, cancel the change.
        new_back = getattr(new_value, back.name, None)
        if (
            new_value
            and new_back
            and self.required
            and new_back is not new_value
        ):
            raise ValueError(
                f"this operation would set the field {back.name} to "
                f"None on {new_back}, which would break model consistency."
            )

    def perform_update(self, model: "Model", old_value: Any, new_value: Any):
        """Update and propagate the update of this field.

        This method is used to propagate updates to linked fields,
        for instance, the value of updated linked fields through relations.

        Args:
            model (Model): the model object to be updated.
            old_value (Any): the old value.
            new_value (Any): the new value already set in the model.

        """
        back = self.__back__
        if old_value:
            old_value._exists = False
            setattr(old_value, back.name, None)
            old_value._exists = True

        old_back = getattr(getattr(model, self.name, None), back.name, None)
        if old_back:
            old_back._exists = False
            setattr(old_back, self.name, None)
            old_back._exists = True

        # Bind the new value.
        if new_value:
            # Modify this side of the relationship.
            model._exists = False
            setattr(model, self.name, new_value)
            model._exists = True

            # Modify the other side of the relationship.
            new_value._exists = False
            setattr(new_value, back.name, model)
            new_value._exists = True

    def validate_delete(self, model: "Model"):
        """Validate whether deleting this field is possible.

        This method should raise an appropriate exception
        if the deletion isn't allowed for any reason.

        Args:
            model (Model): the model object to be updated.

        """
        back = self.__back__
        back_obj = getattr(model, self.name, None)
        if back.required:
            raise ValueError(
                f"cannot delete {model}.{self.name}, since it would "
                f"set {back_obj}.{back.name} to None, while this "
                "field is mandatory"
            )

    def perform_delete(self, model: "Model"):
        """Delete and propagate if necessary.

        This method is used to propagate deletions to linked fields,
        for instance, the value of deleted linked fields through relations.

        Args:
            model (Model): the model object to be updated.

        """
        back = self.__back__
        back_obj = getattr(model, self.name, None)
        if back_obj:
            back_obj._exists = False
            setattr(back_obj, back.name, None)
            back_obj._exists = True



# Copyright (c) 2022, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

"""Helpers to select Pygasus fields."""

from pydantic import BaseModel
from pydantic.fields import SHAPE_SINGLETON

from pygasus.field.base import PygasusField
from pygasus.field.one2one import One2OneField


def update_pygasus_field(field: PygasusField) -> PygasusField:
    """Return a sub-class of PygasusField for this field.

    Args:
        field (ModelField): the model's fieldl itself.

    """
    model = field.__model__
    storage = field.__storage__
    pydantic = field.__field__
    if pydantic.shape == SHAPE_SINGLETON and issubclass(
        pydantic.type_, BaseModel
    ):
        # Check that the back field is a one-relation too.
        back = storage.get_back_field(model, pydantic, pydantic.type_)
        if back.shape == SHAPE_SINGLETON:
            # That's a one-to-one relation.
            new_field = One2OneField(field)
            model.__pygasus__[pydantic.name] = new_field
            setattr(model, pydantic.name, new_field)
            new_field.bind(model, storage)
            return new_field

    return field



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

"""Module containing Sequence to host a list (or set) or data (or models)."""

from collections.abc import MutableSequence
from typing import Any, Generic, List, TypeVar

from pydantic import PrivateAttr
from pydantic.generics import GenericModel

from pygasus.model.helpers import get_primary_keys

model = TypeVar("model")


class Sequence(GenericModel, MutableSequence, Generic[model]):

    """An optionally-sorted sequence of models, behaving like a list.

    Its major role is to behave like a list and to forward modifications
    to the storage system being used.

    """

    models: List[model] = []
    left_model: Any
    right_model: Any
    left_field: Any
    right_field: Any
    parent: Any = None
    _loaded = PrivateAttr()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._loaded = False

    def __contains__(self, model):
        self.load_from_storage()
        return any(model == contained for contained in self.models)

    def __getitem__(self, index: int) -> model:
        return self.models[index]

    def __setitem__(self, index: int, value: model) -> None:
        self.models[index] = value

    def __delitem__(self, index: int) -> None:
        del self.models[index]

    def __len__(self) -> int:
        return len(self.models)

    def __repr__(self) -> str:
        sequence = []
        for element in self.models:
            representation = "<"
            representation += type(element).__name__
            representation += "("
            fields = []
            for name, value in get_primary_keys(element).items():
                fields.append(f"{name}={value!r}")
            representation += ", ".join(fields)
            representation += ")>"
            sequence.append(representation)
        return f"[{', '.join(sequence)}]"

    def __str__(self) -> str:
        sequence = []
        for element in self.models:
            representation = type(element).__name__
            representation += "("
            fields = []
            for name, value in get_primary_keys(element).items():
                fields.append(f"{name}={value!r}")
            representation += ", ".join(fields)
            representation += ")"
            sequence.append(representation)
        return f"[{', '.join(sequence)}]"

    def __eq__(self, other):
        """Compare self to a list."""
        if isinstance(other, Sequence):
            other = other.models

        return len(self.models) == len(other) and all(
            mod1 == mod2 for mod1, mod2 in zip(self.models, other)
        )

    def __iter__(self):
        return iter(self.models)

    def insert(self, index: int, value: model) -> None:
        self.models.insert(index, value)

    def append_new(self, **kwargs) -> model:
        """Append a new model object to the list.

        Args:
            Only keyword arguments are supported.  They're used to create
            the new model, using the model's repository.

        Returns:
            model (model instance): the appended model.

        """
        kwargs[self.right_field.name] = self.parent
        obj = self.right_model.repository.insert_at(len(self), **kwargs)
        self.append(obj)
        return obj

    def insert_new(self, *args, **kwargs) -> model:
        """Insert a new model object to the list.

        Args:
            index (int): the index at which to insert the object.
            Additional keyword arguments are supported.  They're used
            to create the new model, using the model's repository.

        Returns:
            model (model instance): the appended model.

        """
        if len(args) != 1:
            raise ValueError("you must specify only one positional argument")

        index = args[0]
        kwargs[self.right_field.name] = self.parent
        obj = self.right_model.repository.insert_at(index, **kwargs)
        self.insert(index, obj)
        return obj

    def load_from_storage(self):
        """Load the list from the storage system."""
        if self._loaded:
            return

        engine = self.right_model.repository.storage_engine
        self.models[:] = engine.get_related(self)
        self._loaded = True

    class Config:
        underscore_attrs_are_private = True
        copy_on_model_validation = False



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""SQLAlchemy storage engine for Pygasus."""

from datetime import date, datetime
import enum
import operator
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union
from uuid import UUID


try:
    from typing import get_origin
except ImportError:
    from typing_compat import get_origin

from pydantic import EmailStr
from pydantic.fields import SHAPE_LIST
from sqlalchemy import (
    Boolean,
    Column,
    Date,
    DateTime,
    Float,
    ForeignKey,
    Index,
    Integer,
    LargeBinary,
    MetaData,
    Table,
    Text,
    create_engine,
    event,
)
from sqlalchemy.sql import select

from pygasus.field.helpers import update_pygasus_field
from pygasus.model import Field, Model, Sequence
from pygasus.storage.abc import AbstractStorageEngine
from pygasus.storage.sql.query_builder import SQLQueryBuilder

SQL_TYPES = {
    bool: Boolean,
    bytes: LargeBinary,
    date: Date,
    datetime: DateTime,
    float: Float,
    int: Integer,
    str: Text,
    UUID: Text,
    EmailStr: Text,
}


class SQLStorageEngine(AbstractStorageEngine):

    """Storage engine for SQLAlchemy.

    This storage engine only supports Sqlite at the moment.

    """

    def __init__(self):
        super().__init__()
        self.query_builder = SQLQueryBuilder(self)
        self.file_name = None
        self.memory = False
        self.tables = {}
        self.uniques = {}
        self.futures = set()

    def init(
        self,
        file_name: Union[str, Path, None] = None,
        memory: bool = False,
        logging: Union[bool, Callable[[str, Tuple[Any]], None]] = True,
    ):
        """Initialize the storage engine.

        Args:
            file_name (str or Path): the file_name in which the database
                    is stored, or will be stored.  It can be a relative
                    or absolute file name.  If you want to just store
                    in memory, don't specify a file name and just set
                    the `memory` argument to `True`.
            memory (bool): whether to store this database in memory or
                    not?  If `True`, the file name is ignored.
            logging (bool): if True (the default), log SQL queries.

        """
        self.file_name = file_name if not memory else None
        self.memory = memory
        self.logging = logging

        # Connect to the database.
        if memory:
            sql_file_name = ":memory:"
        else:
            if isinstance(file_name, str):
                file_name = Path(file_name)

            assert isinstance(file_name, Path)
            if file_name.absolute():
                sql_file_name = str(file_name)
            else:
                sql_file_name = str(file_name.resolve())
            self.file_name = file_name
        self.engine = create_engine(f"sqlite:///{sql_file_name}")

        # Add a function to override lower, as it only supports
        # ASCII in sqlite3.
        @event.listens_for(self.engine, "connect")
        def setup_lower(dbapi_connection, conn_rec):
            dbapi_connection.create_function("pylower", 1, str.lower)

        # Intercept requests to log them, if set.
        @event.listens_for(self.engine, "before_cursor_execute")
        def log_query(conn, cr, statement, parameters, *_):
            log = self.logging
            log = log if callable(log) else print
            if self.logging:
                log(statement.strip(), parameters)

        self.connection = self.engine.connect()
        self.metadata = MetaData()
        self.tables = {}

    def close(self):
        """Close the connection to the storage engine."""
        self.models = {}

    def destroy(self):
        """Close and destroy the storage engine."""
        self.close()
        if self.file_name:
            self.file_name.unlink()

    def bind(self, models: Optional[Set[Type[Model]]] = None):
        """Bind the speicifed models to this controller.

        If the models aren't specified, then bind every model found
        in the default set.  This is often desirable, although adding
        every model means that one should only import a class in order
        for it to be present.  TalisMUD has a lt of systems of dynamic
        importing, which can make this choice complicated in some
        instances.

        """
        super().bind(models)
        self.metadata.create_all(self.engine)

    def bind_model(self, model: Type[Model], names: Dict[str, Any]):
        """Bind this controller to a specific model.

        Args:
            model (subclass of Model): the model to bind.
            names (dict): the dictionary (name: model) of other models.

        """
        columns = []

        # Browse model fields, creating columsn, indexes and constraints.
        indexes = []
        for name, field in model.__fields__.items():
            model.__pygasus__[name].bind(model, self)
            update_pygasus_field(model.__pygasus__[name])
            pygasus = model.__pygasus__[name]
            o_type = field.outer_type_
            f_type = field.type_
            info = field.field_info
            origin = get_origin(o_type)

            # If it's a custom field, use that instead.
            custom = info.extra.get("custom_class")
            if custom:
                f_type = self.add_custom_field_to_model(custom, model, field)

            if origin is list or o_type is Sequence[f_type]:
                # Analyze the back field.  If it's a list too, an
                # intermediate table should be created.
                field.outer_type_ = Sequence[f_type]
                field.default = Sequence[f_type](
                    left_model=model,
                    left_field=field,
                    right_model=f_type,
                    right_field=self.get_back_field(model, field, f_type),
                )
                continue
            elif isinstance(f_type, type) and issubclass(f_type, Model):
                # If the field is optional (and back isn't) do not add.
                back = self.get_back_field(model, field, f_type)

                # Choose the owner field.
                extra = info.extra
                back_extra = back.field_info.extra
                if extra.get("owner") and not back_extra.get("owner"):
                    extra["owner"] = True
                    back_extra["owner"] = False
                elif not extra.get("owner") and back_extra.get("owner"):
                    extra["owner"] = False
                    back_extra["owner"] = True
                elif pygasus.required and not back.required:
                    extra["owner"] = True
                    back_extra["owner"] = False
                elif not pygasus.required and back.required:
                    extra["owner"] = False
                    back_extra["owner"] = True

                if not extra.get("owner"):
                    continue

                # Create foreign key fields.
                to_name = getattr(
                    f_type.__config__, "model_name", f_type.__name__.lower()
                )
                primary_keys = {
                    name: field
                    for name, field in f_type.__fields__.items()
                    if field.field_info.extra.get("primary_key")
                }

                for pname, pfield in primary_keys.items():
                    columns.append(
                        Column(
                            f"{name}_{pname}",
                            None,
                            ForeignKey(f"{to_name}.{pname}"),
                            nullable=not field.required,
                        )
                    )

                # Add sort options.
                if back.shape == SHAPE_LIST:
                    columns.append(
                        Column(
                            f"{name}__index",
                            Integer,
                            nullable=False,
                        )
                    )
                continue

            # Handle enumerations here.
            if isinstance(f_type, type) and issubclass(f_type, enum.Enum):
                e_type = type(list(f_type)[0].value)
                if not all(
                    isinstance(member.value, e_type) for member in f_type
                ):
                    raise ValueError(
                        f"the enumeration {f_type} contains members 3"
                        "of different types"
                    )
                invalid_member = info.extra.get("invalid_member", "INVALID")
                if isinstance(invalid_member, str):
                    invalid = getattr(f_type, invalid_member, ...)
                else:
                    invalid = invalid_member

                if invalid is ...:
                    raise ValueError(
                        f"field {model.__name__}.{name}: no invalid member "
                        f"is provided for the enumeration {f_type}.  The "
                        "invalid member is required when using enumerations "
                        "to let Pygasus know what to do if the stored "
                        "enumeration member doesn't match the enumeration "
                        "class.  Provide one with "
                        "Field(..., invalid_member=MyEnum.INVALID)"
                    )
                f_type = e_type

            sql_type = SQL_TYPES.get(f_type)
            if sql_type is None:
                raise ValueError(f"unknown type: {name} ({f_type})")
            else:
                extra = info.extra
                primary_key = extra.get("primary_key", False)
                unique = extra.get("unique", False)
                index = extra.get("index", False)
                column = Column(
                    name,
                    sql_type,
                    primary_key=primary_key,
                    nullable=not field.required,
                )
            columns.append(column)

            if unique or index:
                indexes.append(Index(f"idx_{name}", name, unique=unique))

        # Add unique constraints on several columns.
        columns += indexes
        for uniques in getattr(model.__config__, "unique", ()):
            if isinstance(uniques, str):
                uniques = uniques.split()

            if any(name not in model.__fields__ for name in uniques):
                raise ValueError(
                    f"the unique constraint {uniques!r} on model {model} "
                    "is invalid, unknown colomn name"
                )
            columns.append(
                Index(f"idx_{'_'.join(uniques)}", *uniques, unique=True)
            )

        model_name = getattr(
            model.__config__, "model_name", model.__name__.lower()
        )
        if model_name in self.models:
            raise ValueError(
                f"the model {model} of name {model_name!r} conflicts with "
                f"the model {self.models[model_name]} which has the same "
                "name.  Please provide a different name to any of "
                "them, either through setting the 'model_name' "
                "configuration variable in the model or through "
                "changing one of their class names."
            )

        self.models[model_name] = model
        self.tables[model_name] = Table(model_name, self.metadata, *columns)

    def insert(
        self,
        model: Type[Model],
        attrs: Dict[str, Any],
        additional: Optional[Dict[str, Any]] = None,
    ) -> Model:
        """Add a new row for this model.

        The model is used to determine the collection in storage.

        Args:
            model (subclass of Model): the model class.
            attrs (dict): the model attributes.
            additional (opt dict): other attributes, not used in the model.

        Returns:
            obj (Model): the new object.

        """
        model_name = getattr(
            model.__config__, "model_name", model.__name__.lower()
        )
        exclude = set()
        linked = set()
        sql = {}
        for name, field in model.__fields__.items():
            info = field.field_info
            o_type = field.outer_type_
            f_type = field.type_
            pygasus = getattr(model, field.name)
            if o_type is Sequence[f_type]:
                exclude.add(name)
                linked.add(name)
                continue

            custom = self.custom_fields.get((model, field.name))
            if custom:
                value = attrs[name]
                sql[name] = custom.to_storage(value)

            if issubclass(f_type, Model):
                right = attrs.get(name, field.get_default())
                if not isinstance(right, Model):  # Skip here (non valid).
                    exclude.add(name)
                    continue

                primary_keys = {
                    name: field
                    for name, field in f_type.__fields__.items()
                    if field.field_info.extra.get("primary_key")
                }

                for pname, pfield in primary_keys.items():
                    sql[f"{name}_{pname}"] = getattr(right, pname)

                exclude.add(name)
                continue

            if issubclass(f_type, enum.Enum):
                sql[name] = attrs[name].value

            # Convert the field, if necessary.
            method = getattr(
                self, f"{o_type.__name__.lower()}_to_storage", None
            )
            if method:
                sql[name] = method(model, field, attrs[name])

            pk = info.extra.get("primary_key", False)
            if pk and issubclass(f_type, int):
                attrs[name] = 1
                exclude.add(name)

        if additional:
            sql.update(additional)

        # Validate the model object.
        obj = model(**attrs)
        for name, attr in attrs.items():
            pygasus = getattr(model, name)
            pygasus.validate_update(obj, None, attr)

        for name, attr in attrs.items():
            pygasus = getattr(model, name)
            pygasus.perform_update(obj, None, attr)

        # Create a row.
        table = self.tables[model_name]
        insert = table.insert().values(dict(obj.dict(exclude=exclude), **sql))
        result = self.connection.execute(insert)

        primary_keys = iter(result.inserted_primary_key)
        obj._exists = False
        pks = []
        for name, field in model.__fields__.items():
            info = field.field_info
            f_type = field.type_
            pk = info.extra.get("primary_key", False)
            if pk:
                if issubclass(f_type, int):
                    value = next(primary_keys)
                    setattr(obj, name, value)
                else:
                    value = attrs[name]
                pks.append(value)
            elif name in linked:
                # This is a linked field, we need to set the parent.
                getattr(obj, name).parent = obj

        obj._exists = True
        self.cache[(model,) + tuple(pks)] = obj

        for name, field in model.__fields__.items():
            info = field.field_info
            o_type = field.outer_type_
            f_type = field.type_
            pk = info.extra.get("primary_key", False)
            if o_type is Sequence[f_type]:
                obj._exists = False
                setattr(getattr(obj, name), "parent", obj)
                obj._exists = True

        return obj

    def insert_at(
        self,
        model: Type[Model],
        attrs: Dict[str, Any],
        index: int,
        additional: Optional[Dict[str, Any]] = None,
    ) -> Model:
        """Add a new row at this index for this model.

        The model is used to determine the collection in storage.

        Args:
            model (subclass of Model): the model class.
            attrs (dict): the model attributes.
            index (int): the index at which to add this object.
            additional (opt dict): other attributes, not used in the model.

        Returns:
            obj (Model): the new object.

        """
        model_name = getattr(
            model.__config__, "model_name", model.__name__.lower()
        )
        table = self.tables[model_name]
        additional = {}
        for name, field in model.__fields__.items():
            o_type = field.outer_type_
            f_type = field.type_
            if issubclass(f_type, Model) and o_type is not Sequence[f_type]:
                right = attrs[name]
                cmp = operator.le if index < 0 else operator.ge
                where = [cmp(getattr(table.c, f"{name}__index"), index)]
                primary_keys = {
                    name: field
                    for name, field in f_type.__fields__.items()
                    if field.field_info.extra.get("primary_key")
                }

                for pname, pfield in primary_keys.items():
                    where.append(
                        getattr(table.c, f"{name}_{pname}")
                        == getattr(right, pname)
                    )

                # Create and send the query.
                col_name = f"{name}__index"
                col = getattr(table.c, col_name)
                new_value = (col - 1) if index < 0 else (col + 1)
                update = (
                    table.update()
                    .where(*where)
                    .values({getattr(table.c, col_name): new_value})
                )
                self.connection.execute(update)
                additional[col_name] = index
        return self.insert(model, attrs, additional)

    def select(self, model: Type[Model], query):
        """Select one or more model instances with the specified query.

        Args:
            model (subclass of Model): the model object.
            query: the query object by which to filter.

        Returns:
            instances (list of Model): the model isntances or None.

        """
        model_name = getattr(
            model.__config__, "model_name", model.__name__.lower()
        )
        table = self.tables[model_name]
        models = [model]
        columns, tables = self._get_columns_for(model)
        sql = select(*columns)
        for join in tables:
            sql = sql.join(table, isouter=True)
        sql = sql.where(query)

        # Send the query.
        rows = self.connection.execute(sql).fetchall()
        return [
            self._build_objects_from_row(dict(row), models, first=True)
            for row in rows
        ]

    def get(self, model: Type[Model], **kwargs):
        """Get a model instance with the specified arguments.

        Keyword arguments to this method should hold values for
        primary key(s) of the specified model
        (`id` is a common one).

        Args:
            model (subclass of Model): the model object.

        Keyword arguments should contain primary key(s) value(s)
        or other unique fields.

        Returns:
            instance (Model): the model isntance or None.

        Note:
            This method will raise an exception if more than one
            stored instance matches the specified filters.

        """
        # First, check whether this object is in cache.
        pks = []
        model_name = getattr(
            model.__config__, "model_name", model.__name__.lower()
        )
        table = self.tables[model_name]
        models = [model]
        for name, field in model.__fields__.items():
            info = field.field_info
            o_type = field.outer_type_
            f_type = field.type_
            pk = info.extra.get("primary_key", False)
            if pk:
                value = kwargs.get(name, ...)
                if value is not ...:
                    pks.append(value)
            elif issubclass(f_type, Model) and o_type is not Sequence[f_type]:
                # Join up.
                models.append(f_type)

        obj = self.cache.get((model,) + tuple(pks))
        if obj is not None:
            return obj

        columns, tables = self._get_columns_for(model)
        query = select(*columns)
        for join in tables:
            query = query.join(join, isouter=True)
        where = []
        for column, value in kwargs.items():
            method = getattr(
                self, f"{type(value).__name__.lower()}_to_storage", None
            )
            if method:
                value = method(model, model.__fields__[column], value)

            where.append(getattr(table.c, column) == value)
        query = query.where(*where)

        # Send the query.
        rows = self.connection.execute(query).fetchall()
        if len(rows) == 0 or len(rows) < 1:
            return None

        row = dict(rows[0])
        obj = self._build_objects_from_row(row, models, first=True)
        return obj

    def get_related(self, sequence: Sequence) -> List[Model]:
        """Retrieve the related objects from a sequence.

        The sequence is a sequence that represents various objects
        related to the parent model.

        Args:
            sequence (Sequence): the sequence.

        Returns:
            models (list of Model): the models.

        """
        # Create a filter query.
        parent = sequence.parent
        left_model = sequence.left_model
        right_model = sequence.right_model
        pks = {}
        left_model_name = getattr(
            left_model.__config__, "model_name", left_model.__name__.lower()
        )
        right_model_name = getattr(
            right_model.__config__, "model_name", right_model.__name__.lower()
        )
        table = self.tables[right_model_name]
        for name, field in right_model.__fields__.items():
            info = field.field_info
            pk = info.extra.get("primary_key", False)
            if pk:
                value = getattr(parent, name, ...)
                if value is not ...:
                    pks[f"{left_model_name}_{name}"] = value

        columns, tables = self._get_columns_for(right_model)
        query = select(*columns)
        for join in tables:
            query = query.join(join, isouter=True)

        where = []
        for column, value in pks.items():
            where.append(getattr(table.c, column) == value)
        query = query.where(*where)

        # Add ordering.
        query = query.order_by(getattr(table.c, f"{left_model_name}__index"))
        # Send the query.
        rows = self.connection.execute(query).fetchall()
        objs = [
            self._build_objects_from_row(dict(row), [right_model], first=True)
            for row in rows
        ]
        return objs

    def update(
        self,
        model: Type["Model"],
        instance: "Model",
        key: str,
        old_value: Any,
        new_value: Any,
    ):
        """Update an instance attribute.

        Args:
            model (subclass of Model): the model class.
            instance (Model): the model object.
            key (str): the name of the attribute to modify.
            old_value (Any): the value before the modification.
            new_value (Any): the value after the modification.
            check_update (bool): if True (default), check the update.

        """
        model_name = getattr(
            model.__config__, "model_name", model.__name__.lower()
        )
        sql_table = self.tables[model_name]
        sql_primary_keys = []
        for name, field in model.__fields__.items():
            info = field.field_info
            pk = info.extra.get("primary_key", False)
            if pk:
                value = getattr(instance, name)
                if isinstance(value, enum.Enum):
                    value = value.value

                # Handle other field types.
                method = getattr(
                    self, f"{type(value).__name__.lower()}_to_storage", None
                )
                if method:
                    value = method(model, field, value)

                # Handle custom fields.
                custom = self.custom_fields.get((model, field.name))
                if custom:
                    value = custom.to_storage(value)
                sql_primary_keys.append(getattr(sql_table.c, name) == value)

        field = model.__fields__[key]
        # Handles enum field.
        if isinstance(new_value, enum.Enum):
            new_value = new_value.value

        # Handle other field types.
        method = getattr(
            self, f"{type(new_value).__name__.lower()}_to_storage", None
        )
        if method:
            new_value = method(model, field, new_value)

        # Handle custom fields.
        custom = self.custom_fields.get((model, field.name))
        if custom:
            new_value = custom.to_storage(new_value)

        sql_columns = {key: new_value}
        f_type = field.type_
        pygasus = model.__pygasus__[field.name]
        if issubclass(f_type, Model):
            if not field.field_info.extra.get("owner", True) and new_value:
                old_value = getattr(new_value, pygasus.__back__.name, None)
                return self.update(
                    type(new_value),
                    new_value,
                    pygasus.__back__.name,
                    old_value,
                    instance,
                )

            sql_columns = {}
            linked_keys = {
                linked_name: linked_field
                for linked_name, linked_field in f_type.__fields__.items()
                if linked_field.field_info.extra.get("primary_key")
            }

            for pname, pfield in linked_keys.items():
                value = new_value and getattr(new_value, pname) or None
                sql_columns[f"{field.name}_{pname}"] = value

        # Check that this update can be performed.
        pygasus.validate_update(instance, old_value, new_value)
        pygasus.perform_update(instance, old_value, new_value)

        # Send the query.
        update = (
            sql_table.update().where(*sql_primary_keys).values(**sql_columns)
        )
        self.connection.execute(update)

    def delete(self, model: Type[Model], instance: Model):
        """Delete the specified model.

        Args:
            model (subclass of Model): the model class.
            instance (Model): the model instance.

        The model also is removed from cache.

        """
        model_name = getattr(
            model.__config__, "model_name", model.__name__.lower()
        )
        sql_table = self.tables[model_name]
        sql_primary_keys = []
        pks = []
        for name, field in model.__fields__.items():
            info = field.field_info
            pk = info.extra.get("primary_key", False)
            pygasus = model.__pygasus__[name]
            if pk:
                value = getattr(instance, name)
                # Convert other field types.
                method = getattr(
                    self, f"{type(value).__name__.lower()}_to_storage", None
                )
                if method:
                    value = method(model, field, value)
                sql_primary_keys.append(getattr(sql_table.c, name) == value)
                pks.append(value)

            pygasus.validate_delete(instance)

        # Remove this object from cache.
        self.cache.pop((model,) + tuple(pks), None)

        # Send the query.
        delete = sql_table.delete().where(*sql_primary_keys)
        self.connection.execute(delete)

        # Apply other updates if necessary.
        for pygasus in model.__pygasus__.values():
            pygasus.perform_delete(instance)

    def bulk_delete(self, model: Type[Model], query) -> int:
        """Select one or more models with the specified query and delete them.

        Args:
            model (subclass of Model): the model object.
            query: the query object by which to filter.

        Returns:
            number (int): the number of deleted rows.

        """
        model_name = getattr(
            model.__config__, "model_name", model.__name__.lower()
        )
        table = self.tables[model_name]
        pks = []
        for name, field in model.__fields__.items():
            info = field.field_info
            pk = info.extra.get("primary_key", False)
            if pk:
                pks.append(name)

        sql = table.delete().where(query)

        # Perform a select (this will be an additional query, but
        # we need to safely invalidate the cache).
        for obj in self.select(model, query):
            specific_pks = []
            for name in pks:
                value = getattr(obj, name)
                # Convert other field types.
                method = getattr(
                    self, f"{type(value).__name__.lower()}_to_storage", None
                )
                if method:
                    value = method(model, field, value)
                specific_pks.append(value)
            self.cache.pop((model,) + tuple(specific_pks), None)

        # Send the query.
        ret = self.connection.execute(sql)
        return ret.rowcount

    def _get_columns_for(
        self, model: Type[Model], recursion=True
    ) -> Tuple[Column]:
        """Return the relevant columns for the specified model.

        Args:
            model (subclass of Model): the model.

        Returns:
            columns (tuple of Column): the labelled columns for this model.

        """
        model_name = getattr(
            model.__config__, "model_name", model.__name__.lower()
        )
        columns = []
        table = self.tables[model_name]
        tables = []
        for name, field in model.__fields__.items():
            o_type = field.outer_type_
            f_type = field.type_
            if issubclass(f_type, Model) and o_type is not Sequence[f_type]:
                # Join up.
                if recursion:
                    tables.append(
                        self.tables[
                            getattr(
                                f_type.__config__,
                                "model_name",
                                f_type.__name__.lower(),
                            )
                        ]
                    )

                    other_columns, other_tables = self._get_columns_for(
                        f_type, recursion=False
                    )
                    columns.extend(other_columns)
                    tables.extend(other_tables)
            elif o_type is not Sequence[f_type]:
                columns.append(
                    getattr(table.c, name).label(f"{model_name}_{name}")
                )

        return columns, tables

    def _build_objects_from_row(
        self,
        row: Dict[str, Any],
        models: Tuple[Type[Model]],
        first: bool = True,
        done: Optional[Tuple[Type[Model]]] = None,
    ) -> Union[Model, Tuple[Model]]:
        """Create one or more model objects from this row.

        Args:
            row (dict): a dictionary containing row data.
            models (tuple): the tuple of subclasses of models.
            first (bool): if True (the default), return only the first model.
            done (opt): tuple of models that were already processed.

        Returns:
            model or models: the created model(s).

        """
        done = done or ()
        objs = []
        done = list(done)
        customs = []
        for model in models:
            done.append(model)
            model_name = getattr(
                model.__config__, "model_name", model.__name__.lower()
            )

            # Browse the model's fields.
            attrs = {}
            others = []
            pks = []
            not_set = False
            for name, field in model.__fields__.items():
                info = field.field_info
                o_type = field.outer_type_
                f_type = field.type_
                pk = info.extra.get("primary_key", False)
                value = row.get(f"{model_name}_{name}", ...)
                custom = self.custom_fields.get((model, field.name))
                if pk:
                    if value is ... or value is None:
                        not_set = True
                        break

                    pks.append(value)

                    # Convert the field, if necessary.
                    value = row[f"{model_name}_{name}"]
                    method = getattr(
                        self,
                        f"{type(value).__name__.lower()}_from_storage",
                        None,
                    )
                    if method:
                        value = method(model, field, value)

                    attrs[name] = value
                elif custom:
                    value = custom.to_field(value)
                    customs.append((value, name))
                    attrs[name] = value
                elif (
                    isinstance(f_type, type)
                    and issubclass(f_type, Model)
                    and o_type is not Sequence[f_type]
                ):
                    if not field.required:
                        attrs[name] = field.get_default()

                        # Prepare the future.
                        pkeys = tuple(
                            row[f"{model_name}_{pk.name}"]
                            for pk in model.__fields__.values()
                            if pk.field_info.extra.get("primary_key", False)
                        )
                        link_pkeys = tuple(
                            row[f"{name}_{pk.name}"]
                            for pk in f_type.__fields__.values()
                            if pk.field_info.extra.get("primary_key", False)
                        )
                        self.futures.add(
                            (model, pkeys, name, f_type, link_pkeys)
                        )
                    else:
                        obj = self._build_objects_from_row(
                            row, (f_type,), done=tuple(done)
                        )
                        others.append(obj)
                        attrs[name] = obj

                        # Check whether this update would be possible.
                        pygasus = model.__pygasus__[name]
                        pygasus.validate_update(model, None, obj)
                elif issubclass(f_type, enum.Enum):
                    try:
                        value = f_type(value)
                    except ValueError:
                        invalid_key = getattr(
                            model.__config__, "invalid_enum_key", "INVALID"
                        )
                        value = getattr(f_type, invalid_key)
                    finally:
                        attrs[name] = value
                elif o_type is not Sequence[f_type]:
                    value = row[f"{model_name}_{name}"]
                    method = getattr(
                        self,
                        f"{type(value).__name__.lower()}_from_storage",
                        None,
                    )
                    if method:
                        value = method(model, field, value)

                    attrs[name] = value

            # If the object is not complete, don't build it.
            if not_set:
                continue

            obj = self.cache.get((model,) + tuple(pks))
            if obj is not None:
                objs.append(obj)
            else:
                obj = model(**attrs)
                for field, name in customs:
                    field.parent = obj
                    field.field = name
                self.cache[(model,) + tuple(pks)] = obj
                objs.append(obj)
                for name, field in model.__fields__.items():
                    info = field.field_info
                    o_type = field.outer_type_
                    f_type = field.type_
                    pk = info.extra.get("primary_key", False)
                    if o_type is Sequence[f_type]:
                        obj._exists = False
                        col = getattr(obj, name)
                        col.parent = obj
                        obj._exists = True

                # Update Pygasus fields.
                for pygasus in type(obj).__pygasus__.values():
                    value = attrs.get(pygasus.name)
                    pygasus.perform_update(obj, None, value)

            objs.extend(others)

        self._apply_futures()
        if not objs:
            objs = [None]

        return objs[0] if first else objs

    # Type conversions.
    def uuid_from_storage(
        self, model: Model, field: Field, value: str
    ) -> UUID:
        """Convert a UUID hex to a UUID object."""
        return UUID(value)

    def uuid_to_storage(self, model: Model, field: Field, value: UUID) -> str:
        """Convert a UUID to a str representation."""
        return value.hex

    def _apply_futures(self):
        """Try to apply the futures."""
        for model, pks, key, link_model, link_pks in tuple(self.futures):
            obj = self.cache.get((model,) + pks)
            to_link = self.cache.get((link_model,) + link_pks)
            if obj is not None and to_link is not None:
                obj._exists = False
                setattr(obj, key, to_link)
                obj._exists = True
                self.futures.discard((model, pks, key, link_model, link_pks))



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Module containing the abstract class for storage engines."""

from abc import ABCMeta, abstractmethod
from typing import Dict, List, Optional, Set, Type

from pygasus.model import CustomField, Field, Model, Sequence
from pygasus.storage.query_builder import AbstractQueryBuilder


class AbstractStorageEngine(metaclass=ABCMeta):

    """Abstract storage engine.

    The storage engine is here to store, retrieve, create, delete,
    and fetch models from a persistent storage system (like
    a database).  The implementation of such a storage engine
    depends on the storage system being used.

    """

    query_builder: AbstractQueryBuilder = None

    def __init__(self):
        self.models = {}
        self.custom_fields = {}
        self.cache = {}

    def add_custom_field(self, field: Type[CustomField], **kwargs):
        """Add support for a custom field.

        Args:
            field (CustomField): the custom field class.

        Additional keyword arguments can be sent and will be stored
        as options for the custom field.

        """

    def add_custom_field_to_model(
        self,
        custom_field: Type[CustomField],
        model: Model,
        field: Field,
    ) -> CustomField:
        """Add a custom field to a model.

        Args:
            custom_field (CustomField): the custom field to add.
            model (Model): the model to add this field to.
            field (Field): the generic field.

        Returns:
            custom_field (CustomField): the instanciated custom field.

        """
        custom_field = custom_field(self, model, field)
        self.custom_fields[(model, field.name)] = custom_field
        return custom_field.add()

    # Abstract methods
    @abstractmethod
    def init(self):
        """
        Initialize the storage engine.

        Positional and/or keyword arguments can be supported
        in the subclass.  This method is called like a
        "second constructor", where options can be specified and
        connection to the storage engine can be established.
        The connection to the storage engine in particular
        doesn't occur when the object is created, which
        might have side-effects.

        """

    @abstractmethod
    def close(self):
        """Close the connection to the storage engine."""

    @abstractmethod
    def destroy(self):
        """Close and destroy the storage engine."""

    def bind(self, models: Optional[Set[Type[Model]]] = None):
        """Bind the speicifed models to this controller.

        If the models aren't specified, then bind every model found
        in the default set.  This is often desirable, although adding
        every model means that one should only import a class in order
        for it to be present.  TalisMUD has a lt of systems of dynamic
        importing, which can make this choice complicated in some
        instances.

        """
        models = models if models is not None else set()
        names = {model.__name__: model for model in models}
        for model in models:
            model.update_forward_refs(**names)

        for model in models:
            # Generate the model's repository.
            model.repository.storage_engine = self

            # Bind this model to the storage engine.
            self.bind_model(model, names)

    @abstractmethod
    def bind_model(self, model: Type[Model], names: Dict[str, Type[Model]]):
        """Bind this storage engine to a specific model.

        Args:
            model (subclass of Model): the model to bind.
            names (dict): the dictionary (name: model) of other models.

        """

    @abstractmethod
    def insert(self, model: Type[Model], **kwargs):
        """Add a new row for this model.

        The model is used to determine the collection in storage.
        Keyword arguments are used to specify data for this new model.

        Args:
            model (subclass of Model): the model class.
            Additional keyword arguments are expected.

        """

    @abstractmethod
    def get(self, model: Type[Model], **kwargs):
        """Get a model instance with the specified arguments.

        Keyword arguments to this method should hold values for
        primary key(s) of the specified model
        (`id` is a common one).

        Args:
            model (subclass of Model): the model object.

        Keyword arguments should contain primary key(s) value(s)
        or other unique fields.

        Returns:
            instance (Model): the model isntance or None.

        Note:
            This method will raise an exception if more than one
            stored instance matches the specified filters.

        """

    @abstractmethod
    def get_related(self, sequence: Sequence) -> List[Model]:
        """Retrieve the related objects from a collection.

        The collection is a sequence that represents various objects
        related to the parent model.

        Args:
            sequence (Sequence): the collection.

        Returns:
            models (list of Model): the models.

        """

    @abstractmethod
    def update(self, model, instance, key, old_value, new_value):
        """
        Update an instance attribute.

        Args:
            model (subclass of Model): the model class.
            instance (Model: the model object.
            key (str): the name of the attribute to modify.
            old_value (Any): the value before the modification.
            new_value (Any): the value after the modification.

        """

    @abstractmethod
    def delete(self, model: Type[Model], instance: Model):
        """Delete the specified model.

        Args:
            model (subclass of Model): the model class.
            instance (Model): the model instance.

        The model also is removed from cache.

        """

    @abstractmethod
    def bulk_delete(self, model: Type[Model], query) -> int:
        """Select one or more models with the specified query and delete them.

        Args:
            model (subclass of Model): the model object.
            query: the query object by which to filter.

        Returns:
            number (int): the number of deleted rows.

        """

    # Helper methods.
    def get_back_field(self, model, field, left):
        """Return, if found, the bac field of the relationship.

        Args:
            model (subclass of Model): the Model subclass.
            field (Field): the field.
            left (Model): the model linked to the relationship.

        """
        back_name = field.field_info.extra.get("back", "")
        back_fields = (
            {
                b_field
                for b_field in left.__fields__.values()
                if b_field.name == back_name and b_field.type_ is model
            },
            {
                b_field
                for b_field in left.__fields__.values()
                if b_field.type_ is model
            },
        )

        for fields in back_fields:
            if len(fields) == 1:
                return fields.pop()

        # At this point, no unique field was found.
        raise ValueError(
            "no back field could be found for the relationship "
            f"{model.__name__}.{field.name}.  Possible candidates "
            f"are: {fields}.  Perhaps you forgot to add a back reference "
            f"in the {left.__name__} model?"
        )



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

"""Custom field class, to add field support to Pygasus."""

from abc import ABCMeta, abstractmethod
from typing import Any


class CustomField(metaclass=ABCMeta):

    """Abstract class to define a custom field.

    Custom fields should inherit from this class and redefine some
    attributes and methods.  A custom field provides a way
    to store custom information on a model beyond supported fields.

    Attributes to override:
        field_name (str): the field name, a reference.

    Methods to override:
        init(): initializes the field.  This method is called
                when the field is added to a storage engine
                (before it is applied to individual models).
        add(model, field): method called when the custom field
                is added to a model.  The type of the field as supported
                by the storage engine (like `bytes`) should be returned
                at this point, in order for the storage engine
                to know how to store this field.

    """

    def __init__(self, storage_engine, model, field):
        self.storage_engine = storage_engine
        self.model = model
        self.field = field
        self.options = {}

    @abstractmethod
    def add(self) -> Any:
        """Add this field to a model.

        Returns:
            annotation type (Any): the type of field to store.

        """

    @abstractmethod
    def to_storage(self, value: Any) -> Any:
        """Return the value to store in the storage engine.

        Args:
            value (Any): the original value in the field.

        Returns:
            to_store (Any): the value to store.
            It must be of the same type as returned by `add`.

        """

    @abstractmethod
    def to_field(self, value: Any) -> Any:
        """Convert the stored value to the field value.

        Args:
            value (Any): the stored value (same type as returned by `add`).

        Returns:
            to_field (Any): the value to store in the field.
            It must be of the same type as the annotation hint used
            in the model.

        """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pygasus',
 'pygasus.field',
 'pygasus.model',
 'pygasus.model.collections',
 'pygasus.model.decorators',
 'pygasus.storage',
 'pygasus.storage.sql']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.22,<2.0.0',
 'email-validator>=1.1.3,<2.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'typing-compat>=0.1.0,<0.2.0']

setup_kwargs = {
    'name': 'pygasus',
    'version': '0.9.3',
    'description': 'A lightweight ORM built on top of Pydantic',
    'long_description': '# pygasus\n\nA lightweight Sqlite ORM built on top of Pydantic.\n\n## Installation\n\nThe easiest way to install Pygasus is to use `pip`, preferably inside a virtual environment:\n\n    pip install pygasus\n\n`pygasus` supports Python 3.7, 3.8, 3.9 and 3.10.\n\n## Getting started\n\nPygasus allows you to create your models using type annotations.  This model will then be converted to something that can be stored (in a database):\n\n```python\nfrom pygasus import Model\n\nclass User(Model):\n\n    """An user."""\n\n    id: int\n    name: str\n    age: int\n    height: float\n```\n\nYou then need to add this model to Pygasus:\n\n```python\nfrom pygasus.storage import SQLStorageEngine\n\n# The default storage engine is SQLAlchemy, connected to Sqlite3.\nstorage = SQLStorageEngine("test.sql")\n# ... you can also just create the database in memory for testing:\nstorage = SQLStorageEngine(memory=True)\n\n# Then you need to add your models.\nstorage.bind_models({User})\n\n# You can call `bind_models` without any argument, but in this case,\n# make sure to import your `User` class (and your other models)\n# before you do that.\n\n# Create a new user.\nuser = User.repository.create(name="Vincent", age=33, height=5.7)\n# Notice that we don\'t specify the `id`.  This field will be set by\n# the storage engine Pygasus uses.\n\n# At this point, our new user has been created, stored and returned.\nprint(user.id) # 1\nprint(user.name) # Vincent\nuser.age = 21 # Well, why not?\n# This will call an immediate update in the storage.\n\n# Of course we can also query a model:\nqueried = User.repository.get(id=1)\n\n# Due to caching, the queried user will be identical (in terms of reference).\nprint(user is queried) # True\n\n# You can control the cache to remove this behavior.\n```\n\nThis should show you some basic concepts.  To learn more, head over to the documentation.\n\n## Contributing\n\nIf you wish to develop Pygasus, all you need is:\n\n- Docker: getting a working Docker engine isn\'t that complicated on Linux, Windows or Mac OS.\n- Make: getting Make is more complicated, but it\'s not exactly required.  If you\'re on Windows, don\'t despair.  You might rely on Bash itself, it\'s available (and easy to access if you have WSL installed).\n\nPygasus relies on Docker images to make sure it remains usable on the supported Python versions: the idea is that each version will run in a diffferent Docker container, built on a different image.  This process is automated, and you don\'t really need to understand it to use it, but it might be worth diving into the scripts (see the `scripts` directory).\n\n### Initializes the images\n\nOnce you have cloned the `pygasus` repository, you must create the images for the required Python versions.  This is not exactly mandatory, as other scripts will make sure the images exist (and will build them if necessary), but this might make things easier, particularly in case of errors.\n\nPygasus supports four Python versions: 3.7, 3.8, 3.9 and 3.10.  You can initialize them all at once, though it will take time:\n\n    make init\n\nGet yourself a cup of coffee and stare at the process: each image is downloaded and built in turn.  Additional packages are installed.  Pygasus will install additional dependencies, then install Poetry and let it handle other Python dependencies, so that each image uses the same versions in dependencies.\n\nIf you wish to build images individually:\n\n    make init version=3.7\n\nThis will build the image only for Python 3.7 (called pygasus-3.7).  Running each build one at a time makes it easier to spot errors.\n\nOn the other hand, only panics if a build actually fails: it\'s not that uncommon to see warnings and even errors during the build process, but most of the time, this can be ignored... unless the build itself fails (you will be warned if that happens, don\'t worry).\n\n> Curious? Try to look at the images in Docker: `docker images` .  You should see the four `pygasus-VERSION` images along with the `python-VERSION-slim` build (this is the image on which Pygasus images are built).\n\nDon\'t have Make?  If you have Bash, you can still run the command:\n\n    bash scripts/init.sh\n\nInitializing isn\'t a "fire and forget" command.  Should Pygasus dependencies change or critical security bugs be fixed in Python images, it is recommended to run `make init` once more.\n\n### Testing\n\nNow that you have your Pygasus images, you can run the test suite in all versions at once:\n\n    make test\n\nThis will run the test suite in every Python version and display the result on the screen.  This might take a little while, but you should see the progress.  Obviously, if a test fails in a supported Python version, this is a big deal and something to fix for the next commit.  It is recommended to run `make test` before committing anything.\n\nDon\'t have Make?  No worries:\n\n    bash scripts/test.sh\n\nYou can also run the test suite in a specific version, though this is less useful:\n\n    make test version=3.9\n\n### Linter\n\nPygasus uses black and flake8 as linters.  If you have written some code, make sure it\'s properly formatted before committing:\n\n    make check-style\n\nThis won\'t do any modification in your code, this will merely display the errors (if any).  Fix them before committing!\n\nIt\'s also possible to format your code using black:\n\n    make format\n\nThis will actually edit your files in a proper format.  However, notice that flake8 isn\'t involved.  Sometimes, it sees potential errors black will ignore, so make sure `make check-style` is completely empty before committing.\n\nYou don\'t have Make?  No worries:\n\n    bash scripts/check_style.sh\n    bash scripts/format.sh\n\n### Debugging\n\nWhen coding, it can help to actually see what Pygasus does when you execute a line of code.  It is possible to start a shell inside of the container where Pygasus is installed.  You can then access to `pdb` to debug or run some arbitrary code:\n\n    make shell\n\nIf no version is specified, run in the first version (that is, 3.7).  It is assumed that it might be harder to break compatibility that way.  You can specify a version though:\n\n    make shell version=3.10\n\nOnce connected to the container, you can execute commands (including `poetry`, `python` and such).\n\nDon\'t have Make?  No worries:\n\n    bash scripts/shell.sh\n\n### Cleaning up\n\nThe started containers using the previous commands will not be maintained: once you exit them, they will go away.  However, images are kept.  Although this usually doesn\'t pose many problems, these images are heavy.  You can remove images without tags (usually, Pygasus images have got only one tag).\n\n    make clean\n\nThis shouldn\'t be necessary.  Keep an eye on the output of the `docker images` command to see if Docker keeps images it shouldn\'t.\n\n### Updating images\n\nIt bears reminding, once built, Pygasus images aren\'t changed.  If a major security bug is fixed and a new Python image is pushed to Docker hub, you won\'t have it.  If Pygasus dependencies change, you might run into errors.\n\nAt any point, you can force-update the Pygasus images:\n\n    make init\n\nThis will pull the Python image from Docker and rebuild the Pygasus image even if you already have one.\n',
    'author': 'Vincent Le Goff',
    'author_email': 'vincent.legoff.srs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/talismud/pygasus',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Pygasus ORM."""

from pygasus.model import Field, Model  # noqa: F401



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ytranslate nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Abstract query builder.

The query builder is responsible for connecting generic, Pygasus
queries with the selected storage engine.  For instance, when
comparing a field against a value:

```
>>> User.id == 5
<...>
>>>
```

The selected `StorageEngine`, the one that has bound this model,
will return a valid comparator used in queries.  This will speed
up treatment and avoid generic to concrete conversion on the fly.

"""

from abc import ABCMeta, abstractmethod


class AbstractQueryBuilder(metaclass=ABCMeta):

    """Abstract query builder.

    This class is responsible for connecting generic queries
    to the storage engine.  Several methods are provided as abstract.

    Abstract methods:
        eq: provide == comparison.

    """

    def __init__(self, storage_engine):
        self.storage_engine = storage_engine

    @abstractmethod
    def eq(self, field, other):
        """Compare field to other."""

    @abstractmethod
    def ne(self, field, other):
        """Compare field to other."""

    @abstractmethod
    def lt(self, field, other):
        """Compare field to other."""

    @abstractmethod
    def le(self, field, other):
        """Compare field to other."""

    @abstractmethod
    def gt(self, field, other):
        """Compare field to other."""

    @abstractmethod
    def ge(self, field, other):
        """Compare field to other."""

    @abstractmethod
    def is_in(self, field, collection):
        """Filter fields with a value in a collection."""

    @abstractmethod
    def is_not_in(self, field, collection):
        """Filter fields with a value not in a collection."""

    @abstractmethod
    def has(self, field, value):
        """Return models with the field having this value (flag)."""

    @abstractmethod
    def has_not(self, field, collection):
        """Return models without the field having this value (flag)."""



# Copyright (c) 2021, LE GOFF Vincent
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

"""Base class for repositories."""


class Repository:

    storage_engine = None

    def __init__(self, model):
        self.model = model

    def create(self, **data):
        """Create an object.

        Object fields are transmitted as keyword arguments.
        The field must build a complete model object which
        will be validated before a row is written to the
        database.  However, keyword arguments that
        contain information provided by the database
        (such as autoincrement) aren't to be specified.

        Args:
            Keyword arguments to send to the model.

        Example:

            account = repository.create(username="me",
                hashed_password=b"something", ...)

        """
        return self.storage_engine.insert(self.model, data)

    def insert_at(self, *args, **data):
        """Create an object at a given index.

        Object fields are transmitted as keyword arguments.
        The field must build a complete model object which
        will be validated before a row is written to the
        database.  However, keyword arguments that
        contain information provided by the database
        (such as autoincrement) aren't to be specified.

        Args:
            index (int): the index at which to place this model.
            Keyword arguments to send to the model.

        """
        index = args[0]
        return self.storage_engine.insert_at(self.model, data, index)

    def select(self, query):
        """Retrieve model instances based on a query.

        Args:
            The data to be sent to the database.
            query: the query to filter models.

        Examle:
            account = repository.select(Account.username == "me")

        """
        return self.storage_engine.select(self.model, query)

    def get(self, **data):
        """Retrieve a model instance.

        Keyword arguments are used to specify one or more fields
        to identify the object in the database.  Usually,
        these fields are primary keys or unique fields.

        Args:
            The data to be sent to the database.

        Examle :
            account = repository.get(id=5)

        """
        return self.storage_engine.get(self.model, **data)

    def update(self, instance, key, old_value, new_value):
        """Inform of an update."""
        self.storage_engine.update(
            self.model,
            instance,
            key,
            old_value,
            new_value,
        )

    def delete(self, instance):
        """Delete the specified instance.

        This method will remove the instance from the cache as well.

        Args:
            instance (Model): a model instance.

        """
        self.storage_engine.delete(self.model, instance)

    def bulk_delete(self, query) -> int:
        """Delete model instances based on a query.

        Args:
            query: the query to filter models.

        Examle:
            account = repository.bulk_delete(Account.username == "me")

        """
        return self.storage_engine.bulk_delete(self.model, query)
